@@ -5,14 +5,13 @@ import copy
 import logging
 import lxml.etree
 import os
-import os.path
 import pickle
 import posixpath
 import re
 import sys
 import threading
-
-from lxml.etree import XML, XMLSyntaxError
+import Bcfg2.Server
+from Bcfg2.Bcfg2Py3k import ConfigParser
 
 import Bcfg2.Options
 
@@ -51,6 +50,19 @@ info_regex = re.compile( \
     'perms:(\s)*(?P<perms>\w+)|' +
     'sensitive:(\s)*(?P<sensitive>\S+)|')
 
+def bind_info(entry, metadata, infoxml=None, default=default_file_metadata):
+    for attr, val in list(default.items()):
+        entry.set(attr, val)
+    if infoxml:
+        mdata = dict()
+        infoxml.pnode.Match(metadata, mdata, entry=entry)
+        if 'Info' not in mdata:
+            msg = "Failed to set metadata for file %s" % entry.get('name')
+            logger.error(msg)
+            raise PluginExecutionError(msg)
+        for attr, val in list(mdata['Info'][None].items()):
+            entry.set(attr, val)
+
 
 class PluginInitError(Exception):
     """Error raised in cases of Plugin initialization errors."""
@@ -62,7 +74,25 @@ class PluginExecutionError(Exception):
     pass
 
 
-class Plugin(object):
+class Debuggable(object):
+    __rmi__ = ['toggle_debug']
+
+    def __init__(self, name=None):
+        if name is None:
+            name = "%s.%s" % (self.__class__.__module__,
+                              self.__class__.__name__)
+        self.debug_flag = False
+        self.logger = logging.getLogger(name)
+        
+    def toggle_debug(self):
+        self.debug_flag = not self.debug_flag
+
+    def debug_log(self, message, flag=None):
+        if (flag is None and self.debug_flag) or flag:
+            self.logger.error(message)
+
+
+class Plugin(Debuggable):
     """This is the base class for all Bcfg2 Server plugins.
     Several attributes must be defined in the subclass:
     name : the name of the plugin
@@ -77,7 +107,6 @@ class Plugin(object):
     name = 'Plugin'
     __version__ = '$Id$'
     __author__ = 'bcfg-dev@mcs.anl.gov'
-    __rmi__ = ['toggle_debug']
     experimental = False
     deprecated = False
     conflicts = []
@@ -97,16 +126,8 @@ class Plugin(object):
         self.Entries = {}
         self.core = core
         self.data = "%s/%s" % (datastore, self.name)
-        self.logger = logging.getLogger('Bcfg2.Plugins.%s' % (self.name))
         self.running = True
-        self.debug_flag = False
-
-    def toggle_debug(self):
-        self.debug_flag = not self.debug_flag
-
-    def debug_log(self, message, flag=None):
-        if (flag is None) and self.debug_flag or flag:
-            self.logger.error(message)
+        Debuggable.__init__(self, name=self.name)
 
     @classmethod
     def init_repo(cls, repo):
@@ -247,7 +268,9 @@ class ThreadedStatistics(Statistics,
                     if self.terminate.isSet():
                         return False
 
-                self.work_queue.put_nowait((metadata, lxml.etree.fromstring(pdata)))
+                self.work_queue.put_nowait((metadata,
+                                            lxml.etree.XML(pdata,
+                                                           parser=Bcfg2.Server.XMLParser)))
             except Full:
                 self.logger.warning("Queue.Full: Failed to load queue data")
                 break
@@ -266,7 +289,7 @@ class ThreadedStatistics(Statistics,
     def run(self):
         if not self.load():
             return
-        while not self.terminate.isSet():
+        while not self.terminate.isSet() and self.work_queue != None:
             try:
                 (xdata, client) = self.work_queue.get(block=True, timeout=2)
             except Empty:
@@ -276,13 +299,13 @@ class ThreadedStatistics(Statistics,
                 self.logger.error("ThreadedStatistics: %s" % e)
                 continue
             self.handle_statistic(xdata, client)
-        if not self.work_queue.empty():
+        if self.work_queue != None and not self.work_queue.empty():
             self.save()
 
     def process_statistics(self, metadata, data):
         warned = False
         try:
-            self.work_queue.put_nowait((metadata, copy.deepcopy(data)))
+            self.work_queue.put_nowait((metadata, copy.copy(data)))
             warned = False
         except Full:
             if not warned:
@@ -356,7 +379,7 @@ class FileBacked(object):
         object.__init__(self)
         self.data = ''
         self.name = name
-
+        
     def HandleEvent(self, event=None):
         """Read file upon update."""
         if event and event.code2str() not in ['exists', 'changed', 'created']:
@@ -372,6 +395,12 @@ class FileBacked(object):
         """Update local data structures based on current file state"""
         pass
 
+    def __repr__(self):
+        return "%s: %s" % (self.__class__.__name__, str(self))
+
+    def __str__(self):
+        return "%s: %s" % (self.name, self.data)
+
 
 class DirectoryBacked(object):
     """This object is a coherent cache for a filesystem hierarchy of files."""
@@ -556,8 +585,9 @@ class XMLFileBacked(FileBacked):
     def Index(self):
         """Build local data structures."""
         try:
-            self.xdata = XML(self.data)
-        except XMLSyntaxError:
+            self.xdata = lxml.etree.XML(self.data,
+                                        parser=Bcfg2.Server.XMLParser)
+        except lxml.etree.XMLSyntaxError:
             logger.error("Failed to parse %s" % (self.name))
             return
         self.entries = self.xdata.getchildren()
@@ -567,6 +597,9 @@ class XMLFileBacked(FileBacked):
     def __iter__(self):
         return iter(self.entries)
 
+    def __str__(self):
+        return "%s: %s" % (self.name, lxml.etree.tostring(self.xdata))
+
 
 class SingleXMLFileBacked(XMLFileBacked):
     """This object is a coherent cache for an independent XML file."""
@@ -576,10 +609,37 @@ class SingleXMLFileBacked(XMLFileBacked):
         self.fam = fam
         self.fam.AddMonitor(filename, self)
 
+    def _follow_xincludes(self, fname=None, xdata=None):
+        ''' follow xincludes, adding included files to self.extras '''
+        if xdata is None:
+            if fname is None:
+                xdata = self.xdata.getroottree()
+            else:
+                xdata = lxml.etree.parse(fname)
+        included = [el for el in xdata.findall('//%sinclude' %
+                                               Bcfg2.Server.XI_NAMESPACE)]
+        for el in included:
+            name = el.get("href")
+            if name not in self.extras:
+                if name.startswith("/"):
+                    fpath = name
+                else:
+                    fpath = os.path.join(os.path.dirname(self.name), name)
+                if os.path.exists(fpath):
+                    self._follow_xincludes(fname=fpath)
+                    self.add_monitor(fpath, name)
+                else:
+                    msg = "%s: %s does not exist, skipping" % (self.name, name)
+                    if el.findall('./%sfallback' % Bcfg2.Server.XI_NAMESPACE):
+                        self.logger.debug(msg)
+                    else:
+                        self.logger.warning(msg)
+
     def Index(self):
         """Build local data structures."""
         try:
-            self.xdata = lxml.etree.XML(self.data, base_url=self.name)
+            self.xdata = lxml.etree.XML(self.data, base_url=self.name,
+                                        parser=Bcfg2.Server.XMLParser)
         except lxml.etree.XMLSyntaxError:
             err = sys.exc_info()[1]
             logger.error("Failed to parse %s: %s" % (self.name, err))
@@ -636,7 +696,7 @@ class StructFile(XMLFileBacked):
                     rv.extend(self._match(child, metadata))
             return rv
         else:
-            rv = copy.deepcopy(item)
+            rv = copy.copy(item)
             for child in rv.iterchildren():
                 rv.remove(child)
             for child in item.iterchildren():
@@ -745,7 +805,7 @@ class XMLSrc(XMLFileBacked):
             return
         self.items = {}
         try:
-            xdata = lxml.etree.XML(data)
+            xdata = lxml.etree.XML(data, parser=Bcfg2.Server.XMLParser)
         except lxml.etree.XMLSyntaxError:
             logger.error("Failed to parse file %s" % (self.name))
             return
@@ -770,6 +830,9 @@ class XMLSrc(XMLFileBacked):
             self.pnode.Match(metadata, cache[1])
             self.cache = cache
 
+    def __str__(self):
+        return str(self.items)
+
 
 class InfoXML (XMLSrc):
     __node__ = InfoNode
@@ -816,17 +879,16 @@ class PrioDir(Plugin, Generator, XMLDirectoryBacked):
         
     def get_attrs(self, entry, metadata):
         """ get a list of attributes to add to the entry during the bind """
-        for src in list(self.entries.values()):
-            if src.Cache(metadata) == False:
-                self.logger.error("Called before data loaded")
-                raise PluginExecutionError
+        for src in self.entries.values():
+            src.Cache(metadata)
+
         matching = [src for src in list(self.entries.values())
                     if (src.cache and
                         entry.tag in src.cache[1] and
                         self._matches(entry, metadata,
                                       src.cache[1][entry.tag]))]
         if len(matching) == 0:
-            raise PluginExecutionError
+            raise PluginExecutionError('No matching source for entry when retrieving attributes for %s(%s)' % (entry.tag, entry.attrib.get('name')))
         elif len(matching) == 1:
             index = 0
         else:
@@ -845,10 +907,13 @@ class PrioDir(Plugin, Generator, XMLDirectoryBacked):
             if self._matches(entry, metadata, [rname]):
                 data = matching[index].cache[1][entry.tag][rname]
                 break
+        else:
+            # Fall back on __getitem__. Required if override used
+            data = matching[index].cache[1][entry.tag][entry.get('name')]
         if '__text__' in data:
             entry.text = data['__text__']
         if '__children__' in data:
-            [entry.append(copy.deepcopy(item)) for item in data['__children__']]
+            [entry.append(copy.copy(item)) for item in data['__children__']]
 
         return dict([(key, data[key])
                      for key in list(data.keys())
@@ -856,7 +921,6 @@ class PrioDir(Plugin, Generator, XMLDirectoryBacked):
 
 
 # new unified EntrySet backend
-
 class SpecificityError(Exception):
     """Thrown in case of filename parse failure."""
     pass
@@ -972,6 +1036,12 @@ class EntrySet:
             self.entry_init(event)
         else:
             if event.filename not in self.entries:
+                logger.warning("Got %s event for unknown file %s" %
+                               (action, event.filename))
+                if action == 'changed':
+                    # received a bogus changed event; warn, but treat
+                    # it like a created event
+                    self.entry_init(event)
                 return
             if action == 'changed':
                 self.entries[event.filename].handle_event(event)
@@ -1046,18 +1116,7 @@ class EntrySet:
         return cmp(x.specific.prio, y.specific.prio)
 
     def bind_info_to_entry(self, entry, metadata):
-        # first set defaults from global metadata/:info
-        for key in self.metadata:
-            entry.set(key, self.metadata[key])
-        if self.infoxml:
-            mdata = {}
-            self.infoxml.pnode.Match(metadata, mdata, entry=entry)
-            if 'Info' not in mdata:
-                logger.error("Failed to set metadata for file %s" % \
-                             (entry.get('name')))
-                raise PluginExecutionError
-            [entry.attrib.__setitem__(key, value) \
-             for (key, value) in list(mdata['Info'][None].items())]
+        bind_info(entry, metadata, infoxml=self.infoxml, default=self.metadata)
 
     def bind_entry(self, entry, metadata):
         """Return the appropriate interpreted template from the set of available templates."""
@@ -1085,41 +1144,63 @@ class GroupSpool(Plugin, Generator):
         self.AddDirectoryMonitor('')
         self.encoding = core.encoding
 
+    def add_entry(self, event):
+        epath = self.event_path(event)
+        ident = self.event_id(event)
+        if posixpath.isdir(epath):
+            self.AddDirectoryMonitor(epath[len(self.data):])
+        if ident not in self.entries and posixpath.isfile(epath):
+            dirpath = "".join([self.data, ident])
+            self.entries[ident] = self.es_cls(self.filename_pattern,
+                                              dirpath,
+                                              self.es_child_cls,
+                                              self.encoding)
+            self.Entries['Path'][ident] = self.entries[ident].bind_entry
+        if not posixpath.isdir(epath):
+            # do not pass through directory events
+            self.entries[ident].handle_event(event)
+
+    def event_path(self, event):
+        return "".join([self.data, self.handles[event.requestID],
+                        event.filename])
+
+    def event_id(self, event):
+        epath = self.event_path(event)
+        if posixpath.isdir(epath):
+            return self.handles[event.requestID] + event.filename
+        else:
+            return self.handles[event.requestID][:-1]
+
     def HandleEvent(self, event):
-        """Unified FAM event handler for DirShadow."""
+        """Unified FAM event handler for GroupSpool."""
         action = event.code2str()
         if event.filename[0] == '/':
             return
-        epath = "".join([self.data, self.handles[event.requestID],
-                         event.filename])
-        if posixpath.isdir(epath):
-            ident = self.handles[event.requestID] + event.filename
-        else:
-            ident = self.handles[event.requestID][:-1]
+        ident = self.event_id(event)
 
         if action in ['exists', 'created']:
-            if posixpath.isdir(epath):
-                self.AddDirectoryMonitor(epath[len(self.data):])
-            if ident not in self.entries and posixpath.isfile(epath):
-                dirpath = "".join([self.data, ident])
-                self.entries[ident] = self.es_cls(self.filename_pattern,
-                                                  dirpath,
-                                                  self.es_child_cls,
-                                                  self.encoding)
-                self.Entries['Path'][ident] = self.entries[ident].bind_entry
-            if not posixpath.isdir(epath):
-                # do not pass through directory events
+            self.add_entry(event)
+        if action == 'changed':
+            if ident in self.entries:
                 self.entries[ident].handle_event(event)
-        if action == 'changed' and ident in self.entries:
-            self.entries[ident].handle_event(event)
+            else:
+                # got a changed event for a file we didn't know
+                # about. go ahead and process this as a 'created', but
+                # warn
+                self.logger.warning("Got changed event for unknown file %s" %
+                                    ident)
+                self.add_entry(event)
         elif action == 'deleted':
             fbase = self.handles[event.requestID] + event.filename
             if fbase in self.entries:
                 # a directory was deleted
                 del self.entries[fbase]
                 del self.Entries['Path'][fbase]
-            else:
+            elif ident in self.entries:
                 self.entries[ident].handle_event(event)
+            elif ident not in self.entries:
+                self.logger.warning("Got deleted event for unknown file %s" %
+                                    ident)
 
     def AddDirectoryMonitor(self, relative):
         """Add new directory to FAM structures."""
@@ -1132,3 +1213,65 @@ class GroupSpool(Plugin, Generator):
                 return
             reqid = self.core.fam.AddMonitor(name, self)
             self.handles[reqid] = relative
+
+class SimpleConfig(FileBacked,
+                   ConfigParser.SafeConfigParser):
+    ''' a simple plugin config using ConfigParser '''
+    _required = True
+    
+    def __init__(self, plugin):
+        filename = os.path.join(plugin.data, plugin.name.lower() + ".conf")
+        self.plugin = plugin
+        self.fam = self.plugin.core.fam
+        self.read_files = set()
+        Bcfg2.Server.Plugin.FileBacked.__init__(self, filename)
+        ConfigParser.SafeConfigParser.__init__(self)
+
+        if (self._required or
+            (not self._required and os.path.exists(self.name))):
+            self.fam.AddMonitor(self.name, self)
+
+    def Index(self):
+        """ Build local data structures """
+        for section in self.sections():
+            self.remove_section(section)
+        self.read_files.update(self.read(self.name))
+
+    def get(self, section, option, **kwargs):
+        """ convenience method for getting config items """
+        default = None
+        if 'default' in kwargs:
+            default = kwargs['default']
+            del kwargs['default']
+        try:
+            return ConfigParser.SafeConfigParser.get(self, section, option,
+                                                     **kwargs)
+        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
+            if default is not None:
+                return default
+            else:
+                raise
+
+    def getboolean(self, section, option, **kwargs):
+        """ convenience method for getting boolean config items """
+        default = None
+        if 'default' in kwargs:
+            default = kwargs['default']
+            del kwargs['default']
+        try:
+            return ConfigParser.SafeConfigParser.getboolean(self, section,
+                                                            option, **kwargs)
+        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError,
+                ValueError):
+            if default is not None:
+                return default
+            else:
+                raise
+
+    @property
+    def loaded(self):
+        if os.path.exists(self.name):
+            return self.name in self.read_files
+        else:
+            return True
+            
